/* Copyright 1989 GROUPE BULL -- See licence conditions in file COPYRIGHT */
/*****************************************************************************
 ****  Read utility for XPM file format... [ X PixMap ]
 ****
 ****  int    XReadPixmapFile(dpy, d, cmap, filename, w, h, depth, pixmap)
 ****
 ****  original:  Daniel Dardailler - Bull RC (89/02/22) daniel@mirsa.inria.fr
 ****  contributors:
 ****             James Bash - AT&T Bell Labs            jmb@attunix.att.com
 *****************************************************************************
 ****  Version 1.3:  Improved storage mgmt & fatal error handling...
 ****  Version 1.2:  Read support for NAME_mono[] option...
 ****  Version 1.1:  extended chars_per_pixel support... [ Read, Create ]
-------------------------------------------------------------------------------
WHO:	Richard Hess			CORP:	Consilium
TITLE:	Project Manager			VOICE:	[415] 691-6342
      [ User Interface Services ]	USNAIL:	640 Clyde Court
UUCP:	...!uunet!cimshop!rhess			Mountain View, CA 94043
-------------------------------------------------------------------------------
 */

#include <stdio.h>

#ifdef VMS
#include "decw$include:Xlib.h"
#else
#include <X11/Xlib.h>
#endif

#include "xpm.h"      /* PixmapOpenFailed, PixmapSuccess .. MAXPRINTABLE */

static char * getline();
  
/**[ XReadPixmapFile ]********************************************************
 *
 *  Read a Pixmap file in an X Pixmap with specified depth and colormap...
 * (THIS VERSION SUPPORTS : '_mono[]' [ optional ] )
 * (THIS VERSION SUPPORTS : '_chars_per_pixel 1', '_chars_per_pixel 2')
 *****************************************************************************/

int XReadPixmapFile (display,d,
		     colormap,filename,width,height,depth,pixmap)
     Display *display;
     Drawable d;
     Colormap colormap ;
     char *filename;
     unsigned int *width, *height;       /* RETURNED */
     int depth ;   
     Pixmap *pixmap;                     /* RETURNED */
{
  GC Gc = NULL;
  XGCValues xgcv;
  XImage *xim = NULL;

  FILE *fstream = NULL;			/* handle on file  */
  char linebuf[256] ;
  char namebuf[256];
  char type[256];
  char name[256];

  int ncolors ;

  CmapEntry * cmap = NULL;           
  int * Tpixel = NULL;

  char c1, c2, c;
  int red, green, blue ;
  XColor xcolor ;
  int i,j,p,n;
  int cpp;
  int mono = 0;
  
#ifdef DEBUG
#define DEBUGG fprintf
#else
#define DEBUGG
#endif

#undef RETURN
#define RETURN(code, flag) \
  { if (cmap)   free (cmap); \
    if (Tpixel) free (Tpixel); \
    if (Gc)     XFreeGC(display, Gc); \
    if (xim)    XDestroyImage(xim); \
    if (flag && *pixmap) \
      { XFreePixmap(display, *pixmap); \
	*pixmap = NULL; } \
    if (fstream) fclose (fstream); \
    return (code); }

  if ((fstream = fopen(filename, "r")) == NULL) {
    RETURN (PixmapOpenFailed, 0);
  }

  getline(linebuf,fstream);
  if ((sscanf(linebuf, "#define %[^_]%s %d", namebuf, type, &p) != 3) 
      || ((strcmp("_format",type)) && (strcmp("_paxformat",type)))
      || (p != XPM_FORMAT)) {
    RETURN (PixmapFileInvalid, 0);
  }
  else strcpy(name,namebuf);

  getline(linebuf,fstream);
  if ((sscanf(linebuf, "#define %[^_]%s %d", namebuf, type, width) != 3)
      || (strcmp(name,namebuf)) 
      || (strcmp("_width",type))) {
    DEBUGG (stderr,"XPM: bad '#define NAME_width n'\n");
    RETURN (PixmapFileInvalid, 0);
  }

  getline(linebuf,fstream);
  if ((sscanf(linebuf, "#define %[^_]%s %d", namebuf, type, height) != 3)
      || (strcmp(name,namebuf)) 
      || (strcmp("_height",type))) {
    DEBUGG (stderr,"XPM: bad '#define NAME_height n'\n");
    RETURN (PixmapFileInvalid, 0);
  }

  getline(linebuf,fstream);
  if ((sscanf(linebuf, "#define %[^_]%s %d", namebuf, type, &ncolors) != 3)
      || (strcmp(name,namebuf)) 
      || (strcmp("_ncolors",type))) {
    DEBUGG (stderr,"XPM: bad '#define NAME_ncolors n'\n");
    RETURN (PixmapFileInvalid, 0);
  }

  if (ncolors > (MAXPRINTABLE*MAXPRINTABLE)) {
    DEBUGG (stderr,"XPM: Too many different colors, version 1\n");
    RETURN (PixmapFileInvalid, 0);
  }

  getline(linebuf,fstream);
  if ((sscanf(linebuf, "#define %[^_]%s %d", namebuf, type, &cpp) != 3)
      || (strcmp(name,namebuf)) || (cpp < 1) || (cpp > 2)
      || (strcmp("_chars_per_pixel",type))) {
    DEBUGG (stderr,"XPM: bad '#define NAME_chars_per_pixel n' [1][2]\n");
    RETURN (PixmapFileInvalid, 0);
  }

  for (n=0; n<2 ; n++) {
    if (n == 0) {
      getline(linebuf,fstream);
      if ((sscanf(linebuf, "static char * %[^_]%s = {",namebuf,type) != 2)
	  || (strcmp(name,namebuf))
	  || (strcmp("_mono[]",type))) continue;
      if (depth != 1) {
	for (i=0; i<=ncolors ; i++) {
	  getline(linebuf,fstream);
	}
	if (strncmp(linebuf, "} ;",3)) {
	  DEBUGG (stderr,"XPM: missing '} ;' in NAME_mono[]\n");
	  RETURN (PixmapFileInvalid, 0);
	}
	getline(linebuf,fstream);
	continue;
      }
      mono = 1;
    }
    else {
      if (mono) getline(linebuf, fstream);
      if ((sscanf(linebuf, "static char * %[^_]%s = {",namebuf,type) != 2)
	  || (strcmp(name,namebuf))
	  || (strcmp("_colors[]",type))) {
	DEBUGG (stderr,"XPM: bad 'static char * NAME_colors[] = {'\n");
	RETURN (PixmapFileInvalid, 0);
      }
      if (mono) {
	for (i=0; i<=ncolors ; i++) {
	  getline(linebuf,fstream);
	}
	if (strncmp(linebuf, "} ;",3)) {
	  DEBUGG (stderr,"XPM: missing '} ;'\n");
	  RETURN (PixmapFileInvalid, 0);
	}
	continue;
      }
    }

    cmap = (CmapEntry *) malloc(ncolors*sizeof(CmapEntry)) ;
    Tpixel = (int *) malloc(ncolors*sizeof(int)) ;
    if ((cmap == NULL) || (Tpixel == NULL))
      RETURN (PixmapNoMemory, 0) ;
    
    getline(linebuf,fstream);
    for (i=0; i<ncolors ; i++) {
      switch (cpp)
	{
	case 1:
	  if (sscanf(linebuf, "\"%c\" , \"%[^\"]%s",
		     &c1,namebuf,type) != 3) {
	    DEBUGG (stderr,
		    "XPM: bad cmap entry, must be '\"c\" , \"colordef\",'\n"); 
	    RETURN (PixmapFileInvalid, 0);
	  }
	  if (strchr(printable,c1)) {
	    cmap[i].cixel.c1 = c1 ;
	  } else {
	    DEBUGG (stderr,"XPM: bad cixel value, must be printable\n");
	    RETURN (PixmapFileInvalid, 0);
	  }
	  break;
	case 2:
	  if (sscanf(linebuf, "\"%c%c\" , \"%[^\"]%s",
		     &c1,&c2,
		     namebuf,type) != 4) {
	    DEBUGG (stderr,
		   "XPM: bad cmap entry, must be '\"cC\" , \"colordef\",'\n"); 
	    RETURN (PixmapFileInvalid, 0);
	  }
	  if ((strchr(printable,c1)) &&
	      (strchr(printable,c2))) {
	    cmap[i].cixel.c1 = c1 ;
	    cmap[i].cixel.c2 = c2 ;
	  } else {
	    DEBUGG (stderr,"XPM: bad cixel value, must be printable\n");
	    RETURN (PixmapFileInvalid, 0);
	  }
	  break;
	}
      if (!XParseColor(display,colormap,namebuf,&xcolor)) {
	DEBUGG (stderr,
		"XPM: bad colordef specification, #RGB or colorname\n");
	RETURN (PixmapFileInvalid, 0);
      }
      XAllocColor(display,colormap,&xcolor);
      Tpixel[i] = xcolor.pixel ;
      getline(linebuf,fstream);
    }
    if (strncmp(linebuf, "} ;",3)) {
      DEBUGG (stderr,"XPM: missing '} ;'\n");
      RETURN (PixmapFileInvalid, 0);
    }
  }

  getline(linebuf,fstream);
  if ((sscanf(linebuf, "static char * %[^_]%s = {",namebuf,type) != 2)
      || (strcmp(name,namebuf))
      || (strcmp("_pixels[]",type))) {
    DEBUGG (stderr,"XPM: bad 'static char * NAME_pixels[] = {'\n");
    RETURN (PixmapFileInvalid, 0);
  }

  *pixmap = XCreatePixmap(display,d,*width,*height,depth);
  Gc = XCreateGC(display,*pixmap,0,&xgcv);
  xim = XGetImage(display,*pixmap,0,0,*width,*height,AllPlanes,ZPixmap);

  getline(linebuf,fstream);
  j = 0 ;
  while((j < *height) && strncmp(linebuf, "} ;",3))
    {  
      if (strlen(linebuf) < (cpp*(*width)+2)) {
	DEBUGG (stderr,"XPM: bad pixmap line length %d\n",strlen(linebuf));
	RETURN (PixmapFileInvalid, 1);
      }
      switch (cpp)
	{
	case 1:
	  for (i=1; i<=(*width) ; i++)
	    {
	      c1 = linebuf[i] ;
	      for (p = 0 ; p < ncolors ; p++)
		if (cmap[p].cixel.c1 == c1) break ;
	      if (p == ncolors) {
		DEBUGG (stderr,
			"XPM: cixel \"%c\" not in previous cmap\n",c1);
		RETURN (PixmapFileInvalid, 1);
	      }
	      XPutPixel(xim,i-1,j,Tpixel[p]);
	    }
	  break;
	case 2:
	  for (i=1; i<(2*(*width)) ; i+=2)
	    {
	      c1 = linebuf[i] ;
	      c2 = linebuf[i+1] ;
	      for (p = 0 ; p < ncolors ; p++)
		if ((cmap[p].cixel.c1 == c1)&&(cmap[p].cixel.c2 == c2)) break ;
	      if (p == ncolors) {
		DEBUGG (stderr,
			"XPM: cixel \"%c%c\" not in previous cmap\n",c1,c2);
		RETURN (PixmapFileInvalid, 1);
	      }
	      XPutPixel(xim,i/2,j,Tpixel[p]);
	    }
	  break;
	}
      j++ ;
      getline(linebuf,fstream);    
    }

  if (strncmp(linebuf, "} ;",3)) {
    DEBUGG (stderr,"XPM: missing '} ;'\n");
    RETURN (PixmapFileInvalid, 1);
  }

  if (j != *height) {
    DEBUGG (stderr,"XPM: %d too few pixmap line\n", *height - j);
    RETURN (PixmapFileInvalid, 1);
  }
  
  XPutImage(display,*pixmap,Gc,xim,0,0,0,0,*width,*height);

  RETURN (PixmapSuccess, 0);
}


/****[ UTILITIES ]************************************************************
 * following routines are used in XReadPixmapFile() function 
 *****************************************************************************/

/*
 * read the next line and jump blank lines 
 */
static char *
getline(s,pF)
     char * s ;
     FILE * pF ;
{
    s = fgets(s,256,pF);

    while (s) {
	int len = strlen(s);
	if (len && s[len-1] == '\015')
	    s[--len] = '\0';
	if (len==0) s = fgets(s,256,pF);
	else break;
    }
    return(s);
}
	    
/****<eof>********************************************************************/
